#ifndef TREE_H
#define TREE_H

#include <containers/stack.h>
#include <containers/vec_.h>
#include <types.h>

typedef struct TreeNode {
	void const*			data;
	struct TreeNode* 	parent;
	Vec					children;
} TreeNode;

TreeNode* new_TreeNode( void const* data, void (*destroy_func)(void*, void*) );
void delete_TreeNode( TreeNode* self );
void init_TreeNode( TreeNode* self, void const* data, void (*destroy_func)(void*, void*) );
void destroy_TreeNode( TreeNode* self );
TreeNode clone_TreeNode( TreeNode const* self, void* (*clone_func)(void const*) );
void attachChild_TreeNode( TreeNode* self, TreeNode* child );
uint_t getNumChildren_TreeNode( TreeNode const* self );
TreeNode* getParent_TreeNode( TreeNode const* self );
TreeNode* getChild_TreeNode( TreeNode const* self, uint_t i );
void const* getData_TreeNode( TreeNode const* self );

typedef struct Tree {
	TreeNode*	root;
	void*		(*clone_func)(void const*);
	void		(*destroy_func)(void*);
} Tree;

void init_Tree( Tree* self, TreeNode* root, void* (*clone_func)(void const*), void (*destroy_func)(void*) );
void destroy_Tree( Tree* self );
Tree clone_Tree( Tree const* self );
TreeNode* getRoot_Tree( Tree const* self );

typedef struct {
	Stack		stack;
	TreeNode*	node;
	TreeNode*	next;
	uint_t		i;
} TreeNodeIterator;

void init_TreeNodeIterator( TreeNodeIterator* self, TreeNode* node );
void destroy_TreeNodeIterator( TreeNodeIterator* self );
TreeNode* inc_TreeNodeIterator( TreeNodeIterator* self );
void reset_TreeNodeIterator( TreeNodeIterator* self, TreeNode* node );

#endif	/* TREE_H */

