#ifndef VEC__H
#define VEC__H

#include <types.h>

typedef struct {
	uint32_t	num_el;
	uint32_t	max_el;
	size_t		el_size;
	size_t		alignment;
	void*		data_unaligned;
	void*		data;
	void*		end;
	void		(*destroy_func)(void*, void*);
	void*		destroy_func_arg;
} Vec;

Vec* new_Vec( size_t el_size, size_t alignment, void (*destroy_func)(void*, void*), void* destroy_func_arg );
void delete_Vec( Vec* self );
void init_Vec( Vec* self, size_t el_size, size_t alignment, void (*destroy_func)(void*, void*), void* destroy_func_arg );
void destroy_Vec( Vec* self );
Vec clone_Vec( Vec const* self, void* (*clone_func)(void const*) );
void resize_Vec( Vec* self, uint32_t newsize );
void pushBack_Vec( Vec* self, void const* element );
void* getElement_Vec( Vec* self, uint32_t i );
void const* getConstElement_Vec( Vec const* self, uint32_t i );
void* getHead_Vec( Vec* self );
void const* getHeadConst_Vec( Vec const* self );
void* getTail_Vec( Vec* self );
uint32_t getLength_Vec( Vec const* self );
size_t getElementSize_Vec( Vec const* self );
void reset_Vec( Vec* self );

typedef struct {
	size_t	num_elements;
	size_t	current_idx;
	void*	current_element;
	Vec*	vec;
} VecIterator;

VecIterator createIterator_Vec( Vec* self );
void init_VecIterator( VecIterator* self, Vec* vec );
void destroy_VecIterator( VecIterator* self );
void* dereference_VecIterator( VecIterator const* self );
bool next_VecIterator( VecIterator* self );

typedef struct {
	size_t		num_elements;
	size_t		current_idx;
	void const*	current_element;
	Vec const*	vec;
} VecConstIterator;

VecConstIterator createConstIterator_Vec( Vec const* self );
void init_VecConstIterator( VecConstIterator* self, Vec const* vec );
void destroy_VecConstIterator( VecConstIterator* self );
void const* dereference_VecConstIterator( VecConstIterator const* self );
bool next_VecConstIterator( VecConstIterator* self );

#endif	/* VEC__H */
