#ifndef GFXDEVICE_H
#define GFXDEVICE_H

#include <dltrans.h>
#include <types.h>
#include "nativedl.h"
#include "surface.h"

typedef enum {
	SURFACE_COLOR_DRAW,
	SURFACE_COLOR_DISPLAY,
	SURFACE_COLOR_0,
	SURFACE_COLOR_1,
	SURFACE_Z,
	SURFACE_AUX
} SurfaceType;

typedef enum {
	GFXCOLORMODE_RGBA_5551
} GfxColorMode;

typedef struct {
	uint16_t		width;
	uint16_t		height;
	GfxColorMode	colorMode;
} GfxMode;

typedef struct {
	uint32_t	modeIdx;
	OSMesgQueue	viMessageQ;
	OSMesg		viMessages;
	OSMesgQueue	vbiMessageQ;
	OSMesg		vbiMessages;
	OSMesgQueue	rspMessageQ;
	OSMesg		rspMessages;
	OSMesgQueue	rdpMessageQ;
	OSMesg		rdpMessages;
	bool		isRendering;
	DLTrans		dltrans;
	NativeDL	displayList;
	OSTask		tlist;

	/* surfaces */
	Surface*	colorBuffer;		/* surface to DRAW on */
	Surface		colorBuffer0;
	Surface		colorBuffer1;
	Surface		auxBuffer;
	Surface		zBuffer;

	/* RSP stuff. We use KSEG0 virtual adressing, assuming that RSP maps segment 8 to physical address 0x0 */
	u64			dram_stack[SP_DRAM_STACK_SIZE64] __attribute__((aligned (16)));
	u64			yieldBuffer[OS_YIELD_DATA_SIZE/sizeof(u64)];
} GfxDevice;

void init_GfxDevice( GfxDevice* self );
Surface const* getSurface_GfxDevice( GfxDevice const*, SurfaceType type );
void swapBuffers_GfxDevice( GfxDevice* self );
uint32_t getModeIndex_GfxDevice( GfxDevice const* self );
void setMode_GfxDevice( GfxDevice* self, uint32_t mode_idx );
GfxMode const* enumerateModes_GfxDevice( GfxDevice const* self );
void waitForVBI_GfxDevice( GfxDevice const* self );
bool isRendering_GfxDevice( GfxDevice const* self );
void waitForGPU_GfxDevice( GfxDevice const* self );
NativeDL const* getNativeDL_GfxDevice( GfxDevice const* self );
void runGPU_GfxDevice( GfxDevice const* self, DList const* dlist );
float zAsFloat_GfxDevice( GfxDevice const* self, uint32_t zval );

#endif	/* GFXDEVICE_H */
