#ifndef TEXTURE_H
#define TEXTURE_H

#include <types.h>

typedef enum {
	TEXTURE_FMT_CI4,
	TEXTURE_FMT_RGBA16,
	TEXTURE_FMT_RGBA32,
	TEXTURE_FMT_UNKNOWN
} TextureFormat;

typedef struct {
	uint16_t		width;
	uint16_t		height;
	void*			pixels;
	void*			palette;
	uint32_t		pixels_size;
	uint16_t		palette_size;
	TextureFormat	fmt;
} Texture;

/* static initialization */
#define TEXTURE_INIT(w, h, pix, pixsize, pal, palsize, f)	\
{ .width = w, .height = h, .pixels = pix, .palette = pal, .pixels_size = pixsize, .palette_size = palsize, .fmt = f }

/* interface */
void copy_Texture( Texture* self, Texture const* source );
void destroy_Texture( Texture* self );
uint32_t getPixelCount_Texture( Texture const* self );
uint16_t getWidth_Texture( Texture const* self );
uint16_t getHeight_Texture( Texture const* self );
uint8_t getBytesPerPixel_Texture( Texture const* self );
TextureFormat getFormat_Texture( Texture const* self );
void* getPixels_Texture( Texture* self );
void const* getPixelsConst_Texture( Texture const* self );
void const* getPaletteConst_Texture( Texture const* self );

#endif	/* TEXTURE_H */

