#include <abort.h>
#include <sys/region.h>
#include "actmgr/actmgr.h"

void init_ActMgr( ActMgr* actmgr ) {
	init_Vec( &actmgr->actions, sizeof(Action*), ALIGN_DEFAULT, NULL, NULL );
}

void addAction_ActMgr( ActMgr* actmgr, Action* action ) {
	bool found_empty = false;
	Action** current_action;
	VecIterator actions = createIterator_Vec( &actmgr->actions );
	while( (current_action = dereference_VecIterator( &actions )) ) {
		if( NULL == *current_action ) {
			*current_action = action;
			found_empty = true;
			break;
		}

		next_VecIterator( &actions );
	}

	destroy_VecIterator( &actions );

	if( false == found_empty ) {
		pushBack_Vec( &actmgr->actions, &action );
	}
}

void process_ActMgr( ActMgr* actmgr, Time const* time ) {
	static Time const timeInf = ACTMGR_TIME_INFINITY;

	Action** action;
	VecIterator actions = createIterator_Vec( &actmgr->actions );
	while( (action = dereference_VecIterator( &actions )) ) {
		if( TIME_LESS == compare_Time( time, &(*action)->timeStart ) ) {
			next_VecIterator( &actions );
			continue;
		}

		Time time_from_start = calcDuration_Time( &(*action)->timeStart, time );

		if( TIME_EQUAL == compare_Time( &(*action)->timeDuration, &timeInf ) ||
			TIME_LESS == compare_Time( &time_from_start, &(*action)->timeDuration )
		) {
			if( !(*action)->initialized ) {
				(*action)->init_func( &time_from_start, (*action)->arg );
				(*action)->initialized = true;
			}

			if( (*action)->action_func ) {
				(*action)->action_func( &time_from_start, (*action)->arg );
			}
		} else {
			//detachAction_ActMgr( actmgr, action );
			*action = NULL;
		}

		next_VecIterator( &actions );
	}

	destroy_VecIterator( &actions );
}

void detachAction_ActMgr( ActMgr* actmgr, Action* action ) {
	if( actmgr ) {}
	if( action ) {}
	abort();
}

