#include <mesh.h>
#include <layer.h>
#include <sys/region.h>

inline void init_Mesh( Mesh* self ) {
	init_Vec( &self->layers, sizeof( Layer ), ALIGN_DEFAULT, NULL, NULL );
	self->opacity = 1.0f;
	self->attributes = 0;
}

inline void destroy_Mesh( Mesh* self ) {
	VecIterator iterator = createIterator_Vec( &self->layers );

	Layer* layer;
	while( (layer = (Layer*)dereference_VecIterator( &iterator )) ) {
		destroy_Layer( layer );
		next_VecIterator( &iterator );
	}
	destroy_Vec( &self->layers );

	destroy_VecIterator( &iterator );
}

inline void reset_Mesh( Mesh* self ) {
	VecIterator iterator = createIterator_Vec( &self->layers );

	Layer* layer;
	while( (layer = (Layer*)dereference_VecIterator( &iterator )) ) {
		reset_Layer( layer );
		next_VecIterator( &iterator );
	}

	destroy_VecIterator( &iterator );
}

inline void setOpacity_Mesh( Mesh* self, float opacity ) {
	self->opacity = opacity;
}

inline float getOpacity_Mesh( Mesh const* self ) {
	return self->opacity;
}

inline void addLayer_Mesh( Mesh* self, OWN Layer* layer ) {
	pushBack_Vec( &self->layers, layer );
}

inline Vec const* getLayersConst_Mesh( Mesh const* self ) {
	return &self->layers;
}

inline void setAttributes_Mesh( Mesh* self, MeshAttribute attributes ) {
	self->attributes |= attributes;
}

inline MeshAttribute getAttributes_Mesh( Mesh const* self ) {
	return self->attributes;
}
