#include <string.h>
#include <node/xformer.h>

void init_Xformer( Xformer* self, ProcessFunc_Xformer processor ) {
	self->processor	= processor;
	reset_Xformer( self );
}

void destroy_Xformer( Xformer* self ) {
	if( self ) {}
}

inline void process_Xformer( Xformer* self, void* data ) {
	if( self->processor ) {
		self->processor( (struct Xformer*)self, data );
	}
}

inline Matrix* getMatrix_Xformer( Xformer* self ) {
	return &self->mtx;
}

inline Matrix* getResultMatrix_Xformer( Xformer const* self ) {
	return self->resultMtx;
}

inline Coordsys const* getCoordsys_Xformer( Xformer const* self ) {
	return &self->csys;
}

inline void reset_Xformer( Xformer* self ) {
	identity_Matrix( getMatrix_Xformer( self ) );
}

inline void transformLocally_Xformer( Xformer* self, Matrix* mtx ) {
	multiply_Matrix(
		getMatrix_Xformer( self ),
		getMatrix_Xformer( self ),
		mtx
	);
}

void finish_Xformer( Xformer* xf, Xformer* parent_xf, Matrix* resultMtx ) {
	xf->resultMtx = resultMtx;

	if( parent_xf ) {
		multiply_Matrix(
			getResultMatrix_Xformer( xf ),
			getMatrix_Xformer( xf ),
			getResultMatrix_Xformer( parent_xf )
		);
	} else {
		memcpy(
			getResultMatrix_Xformer( xf ),
			getMatrix_Xformer( xf ),
			sizeof( Matrix )
		);
	}

	Matrix const* mtx = getResultMatrix_Xformer( xf );
	xf->csys.origin.x = MATRIX_GET( *mtx, 3, 0 );
	xf->csys.origin.y = MATRIX_GET( *mtx, 3, 1 );
	xf->csys.origin.z = MATRIX_GET( *mtx, 3, 2 );
	xf->csys.n.x = MATRIX_GET( *mtx, 2, 0 );
	xf->csys.n.y = MATRIX_GET( *mtx, 2, 1 );
	xf->csys.n.z = MATRIX_GET( *mtx, 2, 2 );
}

