#include <abort.h>
#include <malloc.h>
#include <string.h>
#include <types.h>
#include <resmgr/resmgr.h>
#include <resmgr/resource.h>
#include <myheader.h>
#include <sys/sys.h>

extern Resource resources[];

static StaticMesh* realloc_mesh( StaticMesh const* inmesh ) {
#define OFS	(0x10001000 - 0x400)
#define ROM_ADR(x) (void const*)( ((uintptr_t)(x) & 0x7FFFFFFF) + OFS )
	StaticMesh* mesh = malloc( sizeof(StaticMesh) );
	if( NULL == mesh ) {
		abort();
	}
	performDMA_Sys( mesh, ROM_ADR( inmesh ), sizeof(StaticMesh) );

	// 1. layers tab
	StaticLayer* layers_copy = malloc( sizeof(StaticLayer) * mesh->num_layers );
	if( NULL == layers_copy ) { abort(); }

	for( uint32_t i=0; i<mesh->num_layers; ++i )
	{
		StaticLayer* layer = &layers_copy[i];
		performDMA_Sys( layer, ROM_ADR( &mesh->layers[i] ), sizeof(StaticLayer) );

		// Gfx
		Gfx* gfx_copy = (Gfx*)malloc( layer->gfx_size );
		if( NULL == gfx_copy ) { abort(); }
		performDMA_Sys( gfx_copy, ROM_ADR( layer->gfx ), layer->gfx_size );
		layer->gfx = gfx_copy;
		// Vtx
		Vtx* vts_copy = (Vtx*)malloc( layer->vts_size );
		if( NULL == vts_copy ) { abort(); }
		performDMA_Sys( vts_copy, ROM_ADR( layer->vts ), layer->vts_size );
		layer->vts_ofs = (uintptr_t)vts_copy - (uintptr_t)layer->vts;
		layer->vts = vts_copy;

	}
	mesh->layers = layers_copy;

	// 2. materials tab
	Material* materials_copy = malloc( sizeof(Material) * mesh->num_materials );
	if( NULL == materials_copy ) { abort(); }

	for( uint32_t i=0; i<mesh->num_materials; ++i )
	{
		performDMA_Sys( &materials_copy[i], ROM_ADR( &mesh->materials[i] ), sizeof(Material) );
	}
	mesh->materials = materials_copy;

	// 3. textures tab
	if( mesh->textures ) {
		Texture* textures_copy = malloc( sizeof(Texture) * mesh->num_textures );
		if( NULL == textures_copy ) { abort(); }
		
		for( uint32_t i=0; i<mesh->num_textures; ++i )
		{
			Texture* texture = &textures_copy[i];
			performDMA_Sys( texture, ROM_ADR( &mesh->textures[i] ), sizeof(Texture) );

			// copy palette
			if( texture->palette_size ) {
				void* palette_copy = malloc( texture->palette_size );
				if( NULL == palette_copy ) { abort(); }
				performDMA_Sys( palette_copy, ROM_ADR( texture->palette ), texture->palette_size );
				texture->palette = palette_copy;
			}
			// copy pixels
			void* pixels_copy = malloc( texture->pixels_size );
			if( NULL == pixels_copy ) { abort(); }
			performDMA_Sys( pixels_copy, ROM_ADR( texture->pixels ), texture->pixels_size );
			texture->pixels = pixels_copy;
		}
		mesh->textures = textures_copy;
	}
	return mesh;
}

void const* getResource_ResMgr( char const* name ) {
	if( NULL == name )
		return NULL;

	Resource* rsc = &resources[0];	
	while( NULL != rsc->name ) {
		if( 0 == strcmp( rsc->name, name ) )
			break;
		rsc++;
	}

	if( RESOURCETYPE_MESH == rsc->type ) {
		return realloc_mesh( rsc->data );
	}

	return rsc->data;
}

