#include <abort.h>
#include "string.h"
#include "malloc.h"

extern uint32_t __heap_start;
extern uint32_t __heap_end;

typedef struct Header {
	struct Header* next;
	struct Header* prev;
	size_t block_size;
	uint32_t pad;		// we need this to keep addresses aligned
} Header;

static size_t calculate_free_space_between( Header* header, Header* next ) {
	if( NULL == next ) { return 0; }

	return (uintptr_t)next - ( (uintptr_t)header + sizeof(Header) + (uintptr_t)header->block_size );
}

void* malloc( size_t size ) {
	size = (size + 0xF) & (~0xF);	// align to 16-bytes

	static bool initialized = false;

	static Header* head;
	static Header* tail;
	if( !initialized ) {
		head = (Header*)(0x80300000 + 320*240*2);	//__heap_start;
		tail = (Header*)(0x80400000 - 8192);		//((Header*)__heap_end) - 1;
	
		head->next = tail;
		head->prev = NULL;
		head->block_size = 0;

		tail->next = NULL;
		tail->prev = head;
		tail->block_size = 0;

		initialized = true;
	}

	Header* current = head;
	while( NULL != current ) {
		if( calculate_free_space_between( current, current->next ) >= size + sizeof(Header) ) {
			Header* new = (Header*)( (uintptr_t)current + sizeof(Header) + current->block_size );
			new->prev = current;
			new->next = current->next;
			new->block_size = size;

			current->next = new;
			new->next->prev = new;

			return (void*)((uintptr_t)new + sizeof(Header));
		}

		current = current->next;
	}

	return NULL;
}

void free( void* ptr ) {
	Header* header = (Header*)( (uintptr_t)ptr - sizeof(Header) );
	header->prev->next = header->next;
	header->next->prev = header->prev;
}

void* realloc( void* ptr, size_t size ) {
	if( !ptr ) {
		return NULL;
	}
	
	void* old_ptr = ptr;
	size_t old_size = ((Header*)( (uintptr_t)ptr - sizeof(Header) ))->block_size;

	ptr = malloc( size );
	if( NULL == ptr ) { abort(); }

	memcpy( ptr, old_ptr, old_size );
	free( old_ptr );

	return ptr;
}

