#include <sys/surface.h>
#include <abort.h>

void init_Surface( Surface* self, PixelFormat pixfmt, void* data,
	uint16_t width, uint16_t height ) {

	self->pixfmt = pixfmt;
	self->data = data;
	self->width = width;
	self->height = height;
}

inline PixelFormat getPixelFormat_Surface( Surface const* self ) {
	return self->pixfmt;
}

inline void* getData_Surface( Surface const* self ) {
	return self->data;
}

inline void const* getValue_Surface( Surface const* self, int16_t x, int16_t y ) {
	if( !coordInBounds_Surface( self, x, y ) ) {
		return NULL;
	}

	switch( self->pixfmt ) {
		case PIXELFMT_RGBA_5551:
		case PIXELFMT_16:
			return &((uint16_t*)self->data)[y * self->width + x];
		default: {
			abort();
		}
	}
}

inline uint16_t getWidth_Surface( Surface const* self ) {
	return self->width;
}

inline uint16_t getHeight_Surface( Surface const* self ) {
	return self->height;
}

inline float getAspectRatio_Surface( Surface const* self ) {
	return (float)getWidth_Surface( self ) / getHeight_Surface( self );
}

inline uint32_t getPixelsNum_Surface( Surface const* self ) {
	return self->width * self->height;
}

inline uint32_t getSize_Surface( Surface const* self ) {
	switch( self->pixfmt ) {
		case PIXELFMT_RGBA_5551:
		case PIXELFMT_16:
			return getPixelsNum_Surface( self ) * 2;
		default:
			return 0;
	}
}

inline bool coordInBounds_Surface( Surface const* self, int16_t x, int16_t y ) {
	return (x >= 0) && (x < self->width) && (y >= 0) && (y < self->height);
}
