#include <abort.h>
#include <malloc.h>
#include <string.h>
#include <sys/sys.h>
#include <ultra64.h>

#define STACKSIZE		0x2000
#define	NUM_PI_MSGS		8
#define	MAINTHREAD_ID	3
#define	MAINTHREAD_PRI	10
#define	VBITHREAD_ID	4
#define	VBITHREAD_PRI	12
#define	SITHREAD_ID		5
#define	SITHREAD_PRI	11
#define AITHREAD_ID		6
#define AITHREAD_PRI	13

#define	FRAMEBUFFER0	((void*)0x80200000)
#define	FRAMEBUFFER1	(void*)((uintptr_t)FRAMEBUFFER0 + BUFFER_W*BUFFER_H*2)
#define	AUXBUFFER		(void*)((uintptr_t)FRAMEBUFFER1 + BUFFER_W*BUFFER_H*2)

#define AUDIOBUFFER					(void*)((uintptr_t)AUXBUFFER + BUFFER_W*BUFFER_H*2)
#define AUDIOBUFFER_SAMPLES_CHNL	(735*2)
#define AUDIOBUFFER_SIZE			(AUDIOBUFFER_SAMPLES_CHNL*2*2)

extern uint32_t __stack_adr;
extern uint32_t __bss_start, __bss_end;
extern void mainproc(void*);

static void idle( void* );

static struct SysState {
	/* thread stuff */
	OSThread	idleThread;
	OSThread	mainThread;
	OSThread	vbiThread;
	OSThread	siThread;
	OSThread	aiThread;
	u64			idleThreadStack[STACKSIZE/sizeof(u64)];
	u64			mainThreadStack[STACKSIZE/sizeof(u64)];
	u64			vbiThreadStack[STACKSIZE/sizeof(u64)];
	u64			siThreadStack[STACKSIZE/sizeof(u64)];
	u64			aiThreadStack[STACKSIZE/sizeof(u64)];

	/* various */
	OSPiHandle*	piHandle;
	int			frameCnt;

	/* messages and message queues */
	OSMesgQueue	siMessageQ;
	OSMesgQueue	aiMessageQ;
	OSMesgQueue	piMessageQ;
	OSMesgQueue	dmaMessageQ;
	OSMesg		siMessages;
	OSMesg		aiMessages;
	OSMesg		piMessages[NUM_PI_MSGS];
	OSMesg		dmaMessages;
	OSIoMesg	dmaIOMessageBuf;

	GfxDevice	gfxDevice;
	AudioDevice	audioDevice;

	/* handlers */
	handler_f	handlerVbi;
	void*		handlerVbiArg;
	handler_f	handlerController;
	void*		handlerControllerArg;
} sysState;

void boot( void ) {
	for(uint32_t* bss = (uint32_t*)__bss_start; bss < (uint32_t*)__bss_end; bss++) {
		*bss = 0;
	}

	osInitialize();
	sysState.piHandle = osCartRomInit();
	osCreateThread(
		&sysState.idleThread,
		1,
		idle,
		(void*)0,
		sysState.idleThreadStack + STACKSIZE/sizeof(u64),
		10
	);
	osStartThread( &sysState.idleThread );
}

static void mainproc_wrapper( void* arg ) {
	init_Sys();

	mainproc( arg );

	raiseError_Sys();
}

static void aiThread( void* arg ) {
	if( arg ) {}

	osCreateMesgQueue(
		&sysState.aiMessageQ,
		&sysState.aiMessages,
		1
	);
	osSetEventMesg(
		OS_EVENT_AI,
		&sysState.aiMessageQ,
		NULL
	);

	uint8_t const* src = getSource_AudioDevice( &sysState.audioDevice );
	uint8_t* dst = (uint8_t*)AUDIOBUFFER;
	if( src ) {
		for(uint32_t i=0; i<AUDIOBUFFER_SIZE; i+=2) {
			*dst++ = *(src+1);
			*dst++ = *(src);
			src += 2;
		}
	}

	osAiSetNextBuffer(
		(uint8_t*)AUDIOBUFFER,
		AUDIOBUFFER_SIZE
	);

	for( int k=0; true; ) {
		//while(!(osAiGetStatus() & AI_STATUS_DMA_BUSY)) {}	// seems like PJ64 can't handle it properly
		osAiSetNextBuffer(
			(uint8_t*)AUDIOBUFFER
				+ (k ? 0 : AUDIOBUFFER_SIZE), AUDIOBUFFER_SIZE
		);

		/* wait till buffer is consumed */
		osRecvMesg(
			&sysState.aiMessageQ,
			NULL,
			OS_MESG_BLOCK
		);

		dst = (uint8_t*)AUDIOBUFFER + (k ? AUDIOBUFFER_SIZE : 0);
		if( src ) {
			for( uint32_t i=0; i<AUDIOBUFFER_SIZE; i+=2 ) {
				*dst++ = *(src+1);
				*dst++ = *(src);
				src += 2;
			}
		} else {
			src = getSource_AudioDevice( &sysState.audioDevice );
		}

		if( (uintptr_t)src >= (uintptr_t)getSource_AudioDevice( &sysState.audioDevice ) + AUDIOBUFFER_SIZE*2*14 ) {
			src = getSource_AudioDevice( &sysState.audioDevice );
		}
			
		osWritebackDCacheAll();

		k ^= 1;
	}
}

static void siThread( void* arg ) {
	if(arg) {}

	OSContPad rdata[ MAXCONTROLLERS ];

	while( 1 ) {
		osRecvMesg(
			&sysState.siMessageQ,
			NULL,
			OS_MESG_BLOCK
		);

		osContGetReadData( rdata );

		if( sysState.handlerController ) {
			sysState.handlerController( &rdata[0], sysState.handlerControllerArg );
		}
	}
}

static void vbiThread( void* arg ) {
	if(arg) {}


	/* wait for vertical blank */
	while( 1 ) {
		osContStartReadData( &sysState.siMessageQ );
		waitForVBI_GfxDevice( &sysState.gfxDevice );
		sysState.frameCnt++;

		/* handle user */
		if( sysState.handlerVbi ) {
			sysState.handlerVbi( NULL, sysState.handlerVbiArg );
		}
		/* forward notification to another queue */
		osJamMesg(
			&sysState.gfxDevice.viMessageQ,
			NULL,
			NULL
		);
	}
}

static void idle( void* arg ) {
	if(arg) {}

	//init_GfxDevice( &sysState.gfxDevice );

	osCreatePiManager( OS_PRIORITY_PIMGR, &sysState.piMessageQ, sysState.piMessages, NUM_PI_MSGS );

	/* main thread */
	osCreateThread(
		&sysState.mainThread,
		MAINTHREAD_ID,
		mainproc_wrapper,
		(void*)0,
		sysState.mainThreadStack + STACKSIZE/sizeof(u64),
		MAINTHREAD_PRI
	);
	osStartThread( &sysState.mainThread );

	osSetThreadPri( 0, OS_PRIORITY_IDLE );

	while( 1 ) {
	}
}

int init_Sys( void ) {
	sysState.frameCnt = 0;

	init_GfxDevice( &sysState.gfxDevice );
	init_AudioDevice( &sysState.audioDevice );

	/* DMA message queue */
	osCreateMesgQueue(
		&sysState.dmaMessageQ,
		&sysState.dmaMessages,
		1
	);

	/* AI stuff */
	osCreateThread(
		&sysState.aiThread,
		AITHREAD_ID,
		aiThread,
		(void*)0,
		sysState.aiThreadStack + STACKSIZE/sizeof(u64),
		AITHREAD_PRI
	);
	osStartThread( &sysState.aiThread );

	/* SI stuff */
	osCreateMesgQueue(
		&sysState.siMessageQ,
		&sysState.siMessages,
		1
	);
	osSetEventMesg(
		OS_EVENT_SI,
		&sysState.siMessageQ,
		NULL
	);

	u8 pattern;
	OSContStatus sdata[MAXCONTROLLERS];
	osContInit(
		&sysState.siMessageQ,
		&pattern,
		sdata
	);

	/* SI thread */
	sysState.handlerController = NULL;
	sysState.handlerControllerArg = NULL;
	osCreateThread(
		&sysState.siThread,
		SITHREAD_ID,
		siThread,
		(void*)0,
		sysState.siThreadStack + STACKSIZE/sizeof(u64),
		SITHREAD_PRI
	);
	osStartThread( &sysState.siThread );

	/* VBI thread */
	sysState.handlerVbi = NULL;
	sysState.handlerVbiArg = NULL;
	osCreateThread(
		&sysState.vbiThread,
		VBITHREAD_ID,
		vbiThread,
		(void*)0,
		sysState.vbiThreadStack + STACKSIZE/sizeof(u64),
		VBITHREAD_PRI
	);
	osStartThread( &sysState.vbiThread );

	return 0;
}

void raiseError_Sys( void ) {
	for( uint32_t i=0; i<(BUFFER_W*BUFFER_H); ++i) {
		((uint8_t*)FRAMEBUFFER0)[i] = 0xFF;
		((uint8_t*)FRAMEBUFFER1)[i] = 0xFF;
	}
}

void registerHandler_Sys( HandlerType type, handler_f handler, void* user_arg ) {
	switch( type ) {
		case HANDLER_VBI: {
			sysState.handlerVbi = handler;
			sysState.handlerVbiArg = user_arg;
			break;
		}
		case HANDLER_CONTROLLER: {
			sysState.handlerController = handler;
			sysState.handlerControllerArg = user_arg;
			break;
		}
		default: {
			break;
		}
	}
}

inline void* getHiMem_Sys( void ) {
	return (void*)(__stack_adr - STACKSIZE);
}

Time getTime_Sys( void ) {
	Time time;
	time.msec = sysState.frameCnt*(1000.0f/60);
	time.sec = time.msec / 1000;
	time.msec %= 1000;

	return time;
}

void performDMA_Sys( void* dest, void const* src, size_t size ) {
	sysState.dmaIOMessageBuf.hdr.pri = OS_MESG_PRI_NORMAL;
	sysState.dmaIOMessageBuf.hdr.retQueue = &sysState.dmaMessageQ;
	sysState.dmaIOMessageBuf.dramAddr = dest;
	sysState.dmaIOMessageBuf.devAddr = (u32)src;
	sysState.dmaIOMessageBuf.size = size;
	osInvalDCache( dest, size );
	osEPiStartDma(
		sysState.piHandle,
		&sysState.dmaIOMessageBuf,
		OS_READ
	);

	OSMesg dummy;
	osRecvMesg( &sysState.dmaMessageQ, &dummy, OS_MESG_BLOCK );
}

GfxDevice* getGfxDevice_Sys( void ) {
	return &sysState.gfxDevice;
}

AudioDevice* getAudioDevice_Sys( void ) {
	return &sysState.audioDevice;
}
