#include <stdlib/abort.h>
#include <actmgr/actmgr.h>
#include <containers/tree.h>
#include <engine.h>
#include <postproc/postproc.h>
#include <resmgr/resmgr.h>
#include <stdlib.h>
#include <sys/sys.h>
#include <sys/region.h>
#include <containers/vec_.h>
#include <string.h>
#include <malloc.h>
#include <mesh.h>
#include "main.h"
#include "img_action.h"
#include "scene_action.h"

#undef POSTPROC_BLUR

static Engine	engine;

typedef struct {
	Player	player;
	float 	angle;
} HandlerArg;

static void handle_vbi( void* sys_arg, void* user_arg ) {
	if( sys_arg ) {}

	if( NULL != user_arg ) {
		*((float*)user_arg) += 1.0f;	// angle
	}
}

static void handle_controller( void* sys_arg, void* user_arg ) {
	static float stx_old,
				 sty_old,
				 ud_old;

	if( (NULL != sys_arg) && (NULL != user_arg) ) {
		OSContPad const* rdata = (OSContPad const*)sys_arg;
		Player* player = (Player*)user_arg;

		player->angle -= 0.04f * stx_old + rdata->stick_x * 0.015f;

		player->dir.x = cosf( M_PI * (player->angle + 90) / 180 );
		//player->dir_y = player->dir.y;
		player->dir.z = -sinf( M_PI * (player->angle + 90) / 180 );

		player->pos.x += 0.0165 * (rdata->stick_y + sty_old) * player->dir.x;
		player->pos.z += 0.0165 * (rdata->stick_y + sty_old) * player->dir.z;
		player->pos.y += 0.03f * ud_old + (rdata->button & U_JPAD && TRUE);
		player->pos.y -= (rdata->button & D_JPAD && TRUE);

		sty_old = sty_old * 0.90f + 0.39f * rdata->stick_y;
		stx_old = 0.87f * stx_old + 0.012f * rdata->stick_x;
		ud_old = ud_old * 0.904f + (rdata->button & U_JPAD && TRUE) - (rdata->button & D_JPAD && TRUE);
	}
}

void process_task( Task const* task ) {
	GfxDevice* gdev = getGfxDevice_Sys();
	waitForGPU_GfxDevice( gdev );

	processTask_Engine( &engine, task );
	runGPU_GfxDevice( gdev, getDisplayList_Engine( &engine ) );
}

void mainproc( void* arg ) {
	if( arg ) {}

	HandlerArg function_args = {
		.player = {
			.pos = VECTOR3F_SET( 0.0f, 130.0f, 200.0f ),	// 400
			.dir = VECTOR3F_SET( 0.0f, 0.0f, 0.0f ),
			.angle = 0.0f	// 60
		},
		.angle = 0.0f
	};

	AudioDevice* audio_dev = getAudioDevice_Sys();
	setSource_AudioDevice( audio_dev, (void*)((uintptr_t)getResource_ResMgr( "sound" ) + 44) );

	init_Engine( &engine );
	registerHandler_Sys( HANDLER_VBI, handle_vbi, &function_args.angle );
	registerHandler_Sys( HANDLER_CONTROLLER, handle_controller, &function_args );

	static Vector3f		clearColor = { 0.5f, 0.3f, 0.8f };
	static float		clearZ = 1.0f;
	static Task 		clearcfbTask = { (taskfptr)clearColorbufF_Engine, &clearColor };
	static Task 		clearzbufTask = { (taskfptr)clearZbufF_Engine, &clearZ };

	/* actions */
	ActMgr actmgr;
	init_ActMgr(&actmgr);

	Time tss = { 0, 1, 0 };
	Time tsd = ACTMGR_TIME_INFINITY;	// { 0, 0, 0 };
	SceneActionArgs scene_args = {
		.engine = &engine,
		.player = &function_args.player,
		.angle = &function_args.angle
	};
	Action sceneAction;
	init_Action( &sceneAction, init_scene_action, handle_scene_action, &scene_args, tss, tsd );
	addAction_ActMgr( &actmgr, &sceneAction );

	#if 0
		Time tid = { 20, 0, 0 };
		ImgActionArgs img_args = {
			.engine = &engine,
			.angle = function_args.angle
		};
		Action imgAction;
		init_Action( &imgAction, init_img_action, handle_img_action, &img_args, tss, tid );
		addAction_ActMgr( &actmgr, &imgAction );
	#endif

	GfxDevice* gdev = getGfxDevice_Sys();
	setZBuffer_Engine( &engine, getSurface_GfxDevice( gdev, SURFACE_Z ) );

	swapBuffers_GfxDevice( gdev );
	while( true ) {
		#ifdef POSTPROC_BLUR
			setRenderTarget_Engine( &engine, getSurface_GfxDevice( gdev, SURFACE_AUX ) );
		#else
			setRenderTarget_Engine( &engine, getSurface_GfxDevice( gdev, SURFACE_COLOR_DRAW ) );
		#endif

		process_task( &clearcfbTask );
		process_task( &clearzbufTask );

		Time current_time = getTime_Sys();
		process_ActMgr( &actmgr, &current_time );
		waitForGPU_GfxDevice( gdev );
		#ifdef POSTPROC_BLUR
			Surface const* src = getSurface_GfxDevice( gdev, SURFACE_AUX );
			Surface const* dest = getSurface_GfxDevice( gdev, SURFACE_COLOR_DRAW );
			postprocBlur( dest, src );
		#endif
		swapBuffers_GfxDevice( gdev );
	}
}

