#include <interpol/interpol.h>
#include <modeler/modeler.h>
#include <resmgr/resmgr.h>
#include <ScalarField.h>
#include <ScalarFieldPolygonizer.h>
#include <stdlib.h>
#include <sys/sys.h>
#include "scene_action.h"

#define BILLBOARDS_NUM		 1
#define FLARE_TEXTURES_NUM	 6

static Material					materials[] = {
	MATERIAL_INIT(
		VECTOR4F_SET( 1.0f, 1.0f, 1.0f, 1.0f ),
		TEXTUREMAPPING_CONST,
		TEXTURE_ID_NONE
	)
};
static ScalarField				scalarField;
static ScalarFieldPolygonizer	polygonizer;
static Vector3f					randoms[BILLBOARDS_NUM];
static Raw_Xformer				root;
static Camera_Xformer			camera_xf;
static StaticMesh_Xformer		worldmesh_xf,
								torusmesh_xf,
								torusmesh2_xf;
static Mesh_Xformer				blobmesh_xf, plane_xf;
static Lux_Xformer				lux_xf_r, lux_xf_b;
static Billboard_Xformer		bb_xf[BILLBOARDS_NUM];
static Flare_Xformer			flare_xf;
static TreeNode					root_node,
								worldmesh_node,
								torusmesh_node,
								torusmesh2_node,
								blobmesh_node,
								lux_r_node,
								lux_b_node,
								camera_node,
								flare_node,
								plane_node,
								bb_node[BILLBOARDS_NUM];
static Tree						scene_tree;
static Vec						flares;
static Mesh						blobmesh;

static float calculate_blobfield( float x, float y, float z, void const* arg ) {
	float g_d = *((float const*)arg);
	float x0 = x-g_d;
	float x1 = x+g_d;
	return 0.5f*( 1.0f/sqrtf( x0*x0 + y*y + z*z ) + 1.0f/sqrtf( x1*x1 + y*y + z*z ) );
}

static void init_meshes( Engine* engine ) {
	worldmesh_xf = createStaticMesh_Engine( engine, (StaticMesh*)getResource_ResMgr( "world" ) );
	init_TreeNode( &worldmesh_node, &worldmesh_xf, NULL );
	attachChild_TreeNode( &root_node, &worldmesh_node );

	torusmesh_xf = createStaticMesh_Engine( engine, (StaticMesh*)getResource_ResMgr( "torus" ) );	// large torus
	torusmesh_xf.static_mesh->opacity = 0.5f;
	init_TreeNode( &torusmesh_node, &torusmesh_xf, NULL );
	attachChild_TreeNode( &root_node, &torusmesh_node );

	torusmesh2_xf = createStaticMesh_Engine( engine, (StaticMesh*)getResource_ResMgr( "torus" ) );	// small torus
	init_TreeNode( &torusmesh2_node, &torusmesh2_xf, NULL );
	attachChild_TreeNode( &torusmesh_node, &torusmesh2_node );

	blobmesh_xf = createMesh_Engine( engine );
	init_TreeNode( &blobmesh_node, &blobmesh_xf, NULL );
	attachChild_TreeNode( &root_node, &blobmesh_node );

	plane_xf = createMesh_Engine( engine );
	*plane_xf.mesh = createPlane_Modeler();
	setAttributes_Mesh( plane_xf.mesh, MESHATTRIBUTE_RECEIVES_SHADOW );
	setOpacity_Mesh( plane_xf.mesh, 0.5f );
	init_TreeNode( &plane_node, &plane_xf, NULL );
	attachChild_TreeNode( &root_node, &plane_node );
}

static void init_lights( Engine* engine ) {
	lux_xf_r = createLux_Engine( engine );
	setColor_Lux( lux_xf_r.lux, 0.9f, 0.0f, 0.0f );
	init_TreeNode( &lux_r_node, &lux_xf_r, NULL );
	attachChild_TreeNode( &torusmesh2_node, &lux_r_node );

	lux_xf_b = createLux_Engine( engine );
	setColor_Lux( lux_xf_b.lux, 0.0f, 0.0f, 0.9f );
	init_TreeNode( &lux_b_node, &lux_xf_b, NULL );
	attachChild_TreeNode( &root_node, &lux_b_node );
}

static void init_billboards( Engine* engine ) {
	for( uint32_t i=0; i<BILLBOARDS_NUM; i++ ) {
		randoms[i].x = 0.004f * rand() * (rand() > 0x4000 ? 1 : -1);
		randoms[i].y = 0.001f * rand() * (rand() > 0x4000 ? 1 : -1) + 50;
		randoms[i].z = 0.004f * rand() * (rand() > 0x4000 ? 1 : -1);

		bb_xf[i] = createBillboard_Engine( engine, (StaticMesh*)getResource_ResMgr( "plane" ) );
		init_TreeNode( &bb_node[i], &bb_xf[i], NULL );
		attachChild_TreeNode( &torusmesh2_node, &bb_node[i] );
	}
}

static void init_flares( Engine* engine ) {
	init_Vec( &flares, sizeof(Pair_StaticMesh_Xformer_Scale), ALIGN_DEFAULT, NULL, NULL );
	for( size_t i=0; i<FLARE_TEXTURES_NUM; ++i ) {
		static float scales[FLARE_TEXTURES_NUM] = { 0.01f, 0.005f, 0.004f, 0.007f, 0.008f, 0.009f };
		Pair_StaticMesh_Xformer_Scale pair = {
			.static_mesh_xf = createStaticMesh_Engine( engine, (StaticMesh const*)getResource_ResMgr( "flare1" ) ),
			.scale = scales[i]
		};
		setTextureCombine_Material( &pair.static_mesh_xf.static_mesh->materials[0], TEXTURECOMBINE_ALPHA );
		pushBack_Vec( &flares, &pair );
	}
	flare_xf = createFlare_Engine( engine, &flares );
	init_TreeNode( &flare_node, &flare_xf, NULL );
	attachChild_TreeNode( &torusmesh2_node, &flare_node );
}

void init_scene_action( Time const* time, void* arg ) {
	if( time ) {}

	SceneActionArgs const* args = (SceneActionArgs const*)arg;
	Engine* engine = args->engine;

	addMaterial_Engine( engine, &materials[0] );

	// scalar field
	init_ScalarField( &scalarField, 9, 9, 9 );
	init_ScalarFieldPolygonizer( &polygonizer );

	// root
	root = createRaw_Engine( engine );
	init_TreeNode( &root_node, &root, NULL );

	// camera
	camera_xf = createCamera_Engine( engine, PERSPECTIVE_CAMERA );
	setParams_PerspCamera( getPersp_Camera( camera_xf.camera ),
		63.74f, 1.33f, 1000.0f
	);
	init_TreeNode( &camera_node, &camera_xf, NULL );
	attachChild_TreeNode( &root_node, &camera_node );

	init_meshes( engine );
	init_lights( engine );
	init_billboards( engine );
	init_flares( engine );

	init_Tree( &scene_tree, &root_node, NULL, NULL );
}

void handle_scene_action( Time const* time, void* arg ) {
	SceneActionArgs* args = (SceneActionArgs*)arg;
	Engine* engine = args->engine;

	static Matrix 		mv;
	static Task 		sceneTask = { (taskfptr)sceneTaskF_Engine, &scene_tree };
	static Task			flareTask = { (taskfptr)flareTaskF_Engine, &scene_tree };

	setCamera_Engine( engine, &camera_xf );

	// camera
	rotate_Matrix( &mv, args->player->angle, 0, 1, 0 );
	transformLocally_Xformer( camera_xf.xf, &mv );
	translate_Matrix( &mv, args->player->pos.x, args->player->pos.y, args->player->pos.z );
	transformLocally_Xformer( camera_xf.xf, &mv );

	// flares
	translate_Matrix( &mv, 0, 0, 0 );
	transformLocally_Xformer( flare_xf.xf, &mv );
	
	// lights
	translate_Matrix( &mv, 0, 1, 0 );
	transformLocally_Xformer( lux_xf_r.xf, &mv );

	// torus
	rotate_Matrix( &mv, *args->angle, 1, 1, 1 );
	transformLocally_Xformer( torusmesh_xf.xf, &mv );
	translate_Matrix( &mv, 0, 160, 0 );
	transformLocally_Xformer( torusmesh_xf.xf, &mv );

	// small torus
	scale_Matrix( &mv, 0.4f, 0.4f, 0.4f );
	transformLocally_Xformer( torusmesh2_xf.xf, &mv );
	Vector3f pos0 = VECTOR3F_SET( 0.0f, -30.0f, 0.0f );
	Vector3f pos1 = VECTOR3F_SET( 0.0f, 30.0f, 0.0f );
	Vector3f pos = calculate_LERP3f( &pos0, &pos1, sinf( 3.2 * time->msec / 1000.f ) );
	translate_Matrix( &mv, pos.x, pos.y, pos.z );
	transformLocally_Xformer( torusmesh2_xf.xf, &mv );

	// billboards
	for( uint32_t i=0; i<BILLBOARDS_NUM; i++ ) {
		translate_Matrix( &mv, randoms[i].x, randoms[i].y, -randoms[i].z );
		transformLocally_Xformer( bb_xf[i].xf, &mv );
	}
	
	// blobs
	float g_d = 0.4*sinf( *args->angle / 100.0f );
	calculate_ScalarField( &scalarField, calculate_blobfield, &g_d );
	blobmesh = process_ScalarFieldPolygonizer( &polygonizer, &scalarField, 3.08f );
	*blobmesh_xf.mesh = blobmesh;
	setAttributes_Mesh( blobmesh_xf.mesh, MESHATTRIBUTE_CASTS_SHADOW );
	translate_Matrix( &mv, 400, 160, 0 );
	transformLocally_Xformer( blobmesh_xf.xf, &mv );

	// plane to cast a shadow on
	translate_Matrix( &mv, 0.0f, 0.0f, -100.0f );
	transformLocally_Xformer( plane_xf.xf, &mv );
	
	// process
	process_task( &sceneTask );
	process_task( &flareTask );

	// cleanup
	destroy_Mesh( &blobmesh );
}

